@@ -149,7 +149,7 @@ typedef struct {
     __u16	starthi;	/* High 16 bits of cluster in FAT32 */
     __u16	time,date,start;/* time, date and first cluster */
     __u32	size;		/* file size (in bytes) */
-} DIR_ENT;
+} __attribute__ ((packed)) DIR_ENT;
 
 typedef struct _dos_file {
     DIR_ENT dir_ent;
@@ -207,6 +207,9 @@ extern void *mem_queue;
 /* return -16 as a number with fs->fat_bits bits */
 #define FAT_EXTD(fs)	(((1 << fs->eff_fat_bits)-1) & ~0xf)
 
+/* marker for files with no 8.3 name */
+#define FAT_NO_83NAME 32
+
 #endif
 
 /* Local Variables: */
